"use strict";

var __assign = this && this.__assign || function () {
  __assign = Object.assign || function (t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
      s = arguments[i];

      for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
    }

    return t;
  };

  return __assign.apply(this, arguments);
};

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.applyTo = exports.INTEGER = void 0;

var types_1 = require("../../libs/types");

var ValueSchemaError_1 = require("../../libs/ValueSchemaError");

var REGEXP_NUMBER = /^\s*[+-]?(\d+(\.\d*)?|\.\d+)\s*$/;
var REGEXP_INTEGER = /^\s*[+-]?\d+\s*$/;
var INTEGER;

(function (INTEGER) {
  INTEGER[INTEGER["NO"] = 0] = "NO";
  INTEGER[INTEGER["YES"] = 1] = "YES";
  INTEGER[INTEGER["FLOOR"] = 2] = "FLOOR";
  INTEGER[INTEGER["FLOOR_RZ"] = 3] = "FLOOR_RZ";
  INTEGER[INTEGER["CEIL"] = 4] = "CEIL";
  INTEGER[INTEGER["CEIL_RI"] = 5] = "CEIL_RI";
  INTEGER[INTEGER["HALF_UP"] = 6] = "HALF_UP";
  INTEGER[INTEGER["HALF_UP_RZ"] = 7] = "HALF_UP_RZ";
  INTEGER[INTEGER["HALF_DOWN"] = 8] = "HALF_DOWN";
  INTEGER[INTEGER["HALF_DOWN_RZ"] = 9] = "HALF_DOWN_RZ";
})(INTEGER = exports.INTEGER || (exports.INTEGER = {}));
/**
 * apply schema
 * @param values input/output values
 * @param options options
 * @param keyStack key stack for error handling
 * @returns applied value
 */


function applyTo(values, options, keyStack) {
  var normalizedOptions = __assign({
    strictType: false,
    acceptsSpecialFormats: false,
    integer: false
  }, options);

  if (types_1.isString(values.output)) {
    if (!checkNumberFormat(normalizedOptions, values.output)) {
      ValueSchemaError_1.ValueSchemaError.raise(ValueSchemaError_1.CAUSE.TYPE, values, keyStack);
    }
  }

  var adjustedValue = toNumber(normalizedOptions, values.output);

  if (adjustedValue === false) {
    ValueSchemaError_1.ValueSchemaError.raise(ValueSchemaError_1.CAUSE.TYPE, values, keyStack);
  }

  values.output = adjustedValue;
  return false;
}

exports.applyTo = applyTo;
/**
 * check the format of value
 * @param options parameters
 * @param value value to check
 * @returns OK/NG
 */

function checkNumberFormat(options, value) {
  var re = getRegExpForNumber(options);

  if (re === null) {
    return true;
  }

  return re.test(value);
}
/**
 * get RegExp pattern for number
 * @param options parameters
 * @returns regular expression pattern
 */


function getRegExpForNumber(options) {
  if (options.acceptsSpecialFormats) {
    return null;
  }

  if (options.integer === true || options.integer === INTEGER.YES) {
    // integer
    return REGEXP_INTEGER;
  } // number


  return REGEXP_NUMBER;
}
/**
 * convert to number
 * @param options parameters
 * @param value value to convert
 * @returns converted value or false(if failed)
 */


function toNumber(options, value) {
  // strict type check
  if (!types_1.isNumber(value) && options.strictType) {
    return false;
  }

  if (!types_1.isScalar(value)) {
    // not a scalar value
    return false;
  }

  var convertedValue = Number(value);

  if (!types_1.isNumber(convertedValue)) {
    // failed to cast
    return false;
  }

  if (options.integer === false || options.integer === INTEGER.NO) {
    return convertedValue;
  } // already integer


  if (types_1.isInteger(convertedValue)) {
    return convertedValue;
  }

  if (options.integer === true || options.integer === INTEGER.YES) {
    // not an integer and no round-off
    return false;
  }

  return round(convertedValue, options.integer);
}
/**
 * round a value
 * @param value value to round
 * @param method round method
 * @returns rounded value
 */


function round(value, method) {
  if (value >= 0) {
    switch (method) {
      case INTEGER.FLOOR:
      case INTEGER.FLOOR_RZ:
        return Math.floor(value);

      case INTEGER.CEIL:
      case INTEGER.CEIL_RI:
        return Math.ceil(value);

      case INTEGER.HALF_UP:
      case INTEGER.HALF_UP_RZ:
        return Math.floor(value + 0.5);

      case INTEGER.HALF_DOWN:
      case INTEGER.HALF_DOWN_RZ:
        return Math.ceil(value - 0.5);
    }
  } else {
    switch (method) {
      case INTEGER.FLOOR:
      case INTEGER.CEIL_RI:
        return Math.floor(value);

      case INTEGER.FLOOR_RZ:
      case INTEGER.CEIL:
        return Math.ceil(value);

      case INTEGER.HALF_UP:
      case INTEGER.HALF_DOWN_RZ:
        return Math.floor(value + 0.5);

      case INTEGER.HALF_UP_RZ:
      case INTEGER.HALF_DOWN:
        return Math.ceil(value - 0.5);
    }
  }

  return false;
}