"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.applyTo = exports.CHECKSUM_ALGORITHM = void 0;

var types_1 = require("../../libs/types");

var ValueSchemaError_1 = require("../../libs/ValueSchemaError");

var CHECKSUM_ALGORITHM;

(function (CHECKSUM_ALGORITHM) {
  CHECKSUM_ALGORITHM["LUHN"] = "luhn";
  CHECKSUM_ALGORITHM["CREDIT_CARD"] = "luhn";
  CHECKSUM_ALGORITHM["MODULUS10_WEIGHT3_1"] = "modulus10/weight3:1";
  CHECKSUM_ALGORITHM["ISBN13"] = "modulus10/weight3:1";
  CHECKSUM_ALGORITHM["EAN"] = "modulus10/weight3:1";
  CHECKSUM_ALGORITHM["JAN"] = "modulus10/weight3:1";
})(CHECKSUM_ALGORITHM = exports.CHECKSUM_ALGORITHM || (exports.CHECKSUM_ALGORITHM = {}));
/**
 * apply schema
 * @param values input/output values
 * @param options options
 * @param keyStack key stack for error handling
 * @returns applied value
 */


function applyTo(values, options, keyStack) {
  if (options.checksum === undefined) {
    return false;
  } // istanbul ignore next


  if (!types_1.isString(values.output)) {
    return false;
  }

  if (check(values.output, options.checksum)) {
    return false;
  }

  return ValueSchemaError_1.ValueSchemaError.raise(ValueSchemaError_1.CAUSE.CHECKSUM, values, keyStack);
}

exports.applyTo = applyTo;
/**
 * check string
 * @param value value to check
 * @param algorithm check algorithm
 * @returns OK/NG
 */

function check(value, algorithm) {
  switch (algorithm) {
    case CHECKSUM_ALGORITHM.LUHN:
      return checkLuhn(value);

    case CHECKSUM_ALGORITHM.MODULUS10_WEIGHT3_1:
      return checkModulus10Weight31(value);

    default:
      return false;
  }
}
/**
 * check by Luhn algorithm (used by credit card)
 * @param value value to check
 * @returns OK/NG
 */


function checkLuhn(value) {
  var length = value.length;
  var sum = 0;

  for (var index = length - 1; index >= 0; index -= 2) {
    // odd columns
    sum += Number(value[index]);
  }

  for (var index = length - 2; index >= 0; index -= 2) {
    // even columns
    var num = Number(value[index]) * 2;
    sum += num % 10;
    sum += Math.floor(num / 10);
  }

  return sum % 10 === 0;
}
/**
 * check by Modulus 10 / Weight 3:1 algorithm (used by ISBN/EAN/JAN)
 * @param value value to check
 * @returns OK/NG
 */


function checkModulus10Weight31(value) {
  var length = value.length;
  var sum = 0;

  for (var index = 0; index < length - 1; index += 2) {
    sum += Number(value[index]);
  }

  for (var index = 1; index < length - 1; index += 2) {
    sum += Number(value[index]) * 3;
  }

  var mod = sum % 10;
  return (10 - mod) % 10 === Number(value[length - 1]);
}